@@ -1,19 +1,21 @@
-# encoding: utf-8
 import bz2
+import io
 import os
 import re
 import shlex
 import warnings
 import zlib
 from collections import OrderedDict
+from typing import IO, AnyStr, Iterable, Tuple
 
 from nrrd.parsers import *
+from nrrd.types import IndexOrder, NRRDFieldMap, NRRDFieldType, NRRDHeader
 
 # Older versions of Python had issues when uncompressed data was larger than 4GB (2^32). This should be fixed in latest
 # version of Python 2.7 and all versions of Python 3. The fix for this issue is to read the data in smaller chunks.
 # Chunk size is set to be large at 1GB to improve performance. If issues arise decompressing larger files, try to reduce
 # this value
-_READ_CHUNKSIZE = 2 ** 32
+_READ_CHUNKSIZE: int = 2 ** 32
 
 _NRRD_REQUIRED_FIELDS = ['dimension', 'type', 'encoding', 'sizes']
 
@@ -84,7 +86,7 @@ _TYPEMAP_NRRD2NUMPY = {
 }
 
 
-def _get_field_type(field, custom_field_map):
+def _get_field_type(field: str, custom_field_map: Optional[NRRDFieldMap]) -> NRRDFieldType:
     if field in ['dimension', 'lineskip', 'line skip', 'byteskip', 'byte skip', 'space dimension']:
         return 'int'
     elif field in ['min', 'max', 'oldmin', 'old min', 'oldmax', 'old max']:
@@ -99,7 +101,7 @@ def _get_field_type(field, custom_field_map):
         return 'string list'
     elif field in ['labels', 'units', 'space units']:
         return 'quoted string list'
-    # No int vector fields as of now
+    # No int vector fields yet
     # elif field in []:
     #     return 'int vector'
     elif field in ['space origin']:
@@ -116,7 +118,7 @@ def _get_field_type(field, custom_field_map):
         return 'string'
 
 
-def _parse_field_value(value, field_type):
+def _parse_field_value(value: str, field_type: NRRDFieldType) -> Any:
     if field_type == 'int':
         return int(value)
     elif field_type == 'double':
@@ -143,31 +145,31 @@ def _parse_field_value(value, field_type):
         # for none rows. NaN is only valid for floating point numbers
         return parse_optional_matrix(value)
     else:
-        raise NRRDError('Invalid field type given: %s' % field_type)
+        raise NRRDError(f'Invalid field type given: {field_type}')
 
 
-def _determine_datatype(fields):
+def _determine_datatype(header: NRRDHeader) -> np.dtype:
     """Determine the numpy dtype of the data."""
 
     # Convert the NRRD type string identifier into a NumPy string identifier using a map
-    np_typestring = _TYPEMAP_NRRD2NUMPY[fields['type']]
+    np_typestring = _TYPEMAP_NRRD2NUMPY[header['type']]
 
     # This is only added if the datatype has more than one byte and is not using ASCII encoding
     # Note: Endian is not required for ASCII encoding
-    if np.dtype(np_typestring).itemsize > 1 and fields['encoding'] not in ['ASCII', 'ascii', 'text', 'txt']:
-        if 'endian' not in fields:
-            raise NRRDError('Header is missing required field: "endian".')
-        elif fields['endian'] == 'big':
+    if np.dtype(np_typestring).itemsize > 1 and header['encoding'] not in ['ASCII', 'ascii', 'text', 'txt']:
+        if 'endian' not in header:
+            raise NRRDError('Header is missing required field: endian')
+        elif header['endian'] == 'big':
             np_typestring = '>' + np_typestring
-        elif fields['endian'] == 'little':
+        elif header['endian'] == 'little':
             np_typestring = '<' + np_typestring
         else:
-            raise NRRDError('Invalid endian value in header: "%s"' % fields['endian'])
+            raise NRRDError(f'Invalid endian value in header: {header["endian"]}')
 
     return np.dtype(np_typestring)
 
 
-def _validate_magic_line(line):
+def _validate_magic_line(line: str) -> int:
     """For NRRD files, the first four characters are always "NRRD", and
     remaining characters give information about the file format version
 
@@ -189,15 +191,15 @@ def _validate_magic_line(line):
     try:
         version = int(line[4:])
         if version > 5:
-            raise NRRDError('Unsupported NRRD file version (version: %i). This library only supports v%i and below.'
-                            % (version, 5))
+            raise NRRDError(f'Unsupported NRRD file version (version: {version}). This library only supports v5 '
+                            'and below.')
     except ValueError:
-        raise NRRDError('Invalid NRRD magic line: %s' % line)
+        raise NRRDError(f'Invalid NRRD magic line: {line}')
 
     return len(line)
 
 
-def read_header(file, custom_field_map=None):
+def read_header(file: Union[str, Iterable[AnyStr]], custom_field_map: Optional[NRRDFieldMap] = None) -> NRRDHeader:
     """Read contents of header and parse values from :obj:`file`
 
     :obj:`file` can be a filename indicating where the NRRD header is located or a string iterator object. If a
@@ -207,7 +209,7 @@ def read_header(file, custom_field_map=None):
     :obj:`file` is a file object, it must be opened with the binary flag ('b') on platforms where that makes a
     difference, such as Windows.
 
-    See :ref:`user-guide:Reading NRRD files` for more information on reading NRRD files.
+    See :ref:`background/how-to-use:reading nrrd files` for more information on reading NRRD files.
 
     Parameters
     ----------
@@ -231,8 +233,7 @@ def read_header(file, custom_field_map=None):
     # file handle. Since read function uses a filename, it is easy to think read_header is the same syntax.
     if isinstance(file, str) and file.count('\n') == 0:
         with open(file, 'rb') as fh:
-            header = read_header(fh, custom_field_map)
-            return header
+            return read_header(fh, custom_field_map)
 
     # Collect number of bytes in the file header (for seeking below)
     header_size = 0
@@ -280,14 +281,12 @@ def read_header(file, custom_field_map=None):
 
         # Check if the field has been added already
         if field in header.keys():
-            dup_message = "Duplicate header field: '%s'" % str(field)
-
             if not ALLOW_DUPLICATE_FIELD:
-                raise NRRDError(dup_message)
+                raise NRRDError(f'Duplicate header field: {field}')
+            else:
+                warnings.warn(f'Duplicate header field: {field}')
 
-            warnings.warn(dup_message)
-
-        # Get the datatype of the field based on it's field name and custom field map
+        # Get the datatype of the field based on its field name and custom field map
         field_type = _get_field_type(field, custom_field_map)
 
         # Parse the field value using the datatype retrieved
@@ -302,7 +301,8 @@ def read_header(file, custom_field_map=None):
     return header
 
 
-def read_data(header, fh=None, filename=None, index_order='F'):
+def read_data(header: NRRDHeader, fh: Optional[IO] = None, filename: Optional[str] = None,
+              index_order: IndexOrder = 'F') -> npt.NDArray:
     """Read data from file into :class:`numpy.ndarray`
 
     The two parameters :obj:`fh` and :obj:`filename` are optional depending on the parameters but it never hurts to
@@ -310,7 +310,7 @@ def read_data(header, fh=None, filename=None, index_order='F'):
     the NRRD data is detached from the header, then the :obj:`filename` parameter is required to obtain the absolute
     path to the data file.
 
-    See :ref:`user-guide:Reading NRRD files` for more information on reading NRRD files.
+    See :ref:`background/how-to-use:reading nrrd files` for more information on reading NRRD files.
 
     Parameters
     ----------
@@ -322,9 +322,9 @@ def read_data(header, fh=None, filename=None, index_order='F'):
         Filename of the header file. Only necessary if data is detached from the header. This is used to get the
         absolute data path.
     index_order : {'C', 'F'}, optional
-        Specifies the index order of the resulting data array. Either 'C' (C-order) where the dimensions are ordered from
-        slowest-varying to fastest-varying (e.g. (z, y, x)), or 'F' (Fortran-order) where the dimensions are ordered
-        from fastest-varying to slowest-varying (e.g. (x, y, z)).
+        Specifies the index order of the resulting data array. Either 'C' (C-order) where the dimensions are ordered
+        from slowest-varying to fastest-varying (e.g. (z, y, x)), or 'F' (Fortran-order) where the dimensions are
+        ordered from fastest-varying to slowest-varying (e.g. (x, y, z)).
 
     Returns
     -------
@@ -342,11 +342,11 @@ def read_data(header, fh=None, filename=None, index_order='F'):
     # Check that the required fields are in the header
     for field in _NRRD_REQUIRED_FIELDS:
         if field not in header:
-            raise NRRDError('Header is missing required field: "%s".' % field)
+            raise NRRDError(f'Header is missing required field: {field}')
 
     if header['dimension'] != len(header['sizes']):
-        raise NRRDError('Number of elements in sizes does not match dimension. Dimension: %i, len(sizes): %i' % (
-            header['dimension'], len(header['sizes'])))
+        raise NRRDError(f'Number of elements in sizes does not match dimension. Dimension: {header["dimension"]}, '
+                        f'len(sizes): {len(header["sizes"])}')
 
     # Determine the data type from the header
     dtype = _determine_datatype(header)
@@ -403,9 +403,16 @@ def read_data(header, fh=None, filename=None, index_order='F'):
 
     # If a compression encoding is used, then byte skip AFTER decompressing
     if header['encoding'] == 'raw':
-        data = np.fromfile(fh, dtype)
+        if isinstance(fh, io.BytesIO):
+            raw_data = bytearray(fh.read(total_data_points * dtype.itemsize))
+            data = np.frombuffer(raw_data, dtype)
+        else:
+            data = np.fromfile(fh, dtype)
     elif header['encoding'] in ['ASCII', 'ascii', 'text', 'txt']:
-        data = np.fromfile(fh, dtype, sep=' ')
+        if isinstance(fh, io.BytesIO):
+            data = np.fromstring(fh.read(), dtype, sep=' ')
+        else:
+            data = np.fromfile(fh, dtype, sep=' ')
     else:
         # Handle compressed data now
         # Construct the decompression object based on encoding
@@ -418,12 +425,12 @@ def read_data(header, fh=None, filename=None, index_order='F'):
             # to close it for us
             fh.close()
 
-            raise NRRDError('Unsupported encoding: "%s"' % header['encoding'])
+            raise NRRDError(f'Unsupported encoding: {header["encoding"]}')
 
         # Loop through the file and read a chunk at a time (see _READ_CHUNKSIZE why it is read in chunks)
         decompressed_data = bytearray()
 
-        # Read all of the remaining data from the file
+        # Read all the remaining data from the file
         # Obtain the length of the compressed data since we will be using it repeatedly, more efficient
         compressed_data = fh.read()
         compressed_data_len = len(compressed_data)
@@ -453,8 +460,8 @@ def read_data(header, fh=None, filename=None, index_order='F'):
     fh.close()
 
     if total_data_points != data.size:
-        raise NRRDError('Size of the data does not equal the product of all the dimensions: {0}-{1}={2}'
-                        .format(total_data_points, data.size, total_data_points - data.size))
+        raise NRRDError(f'Size of the data does not equal the product of all the dimensions: '
+                        f'{total_data_points}-{data.size}={total_data_points - data.size}')
 
     # In the NRRD header, the fields are specified in Fortran order, i.e, the first index is the one that changes
     # fastest and last index changes slowest. This needs to be taken into consideration since numpy uses C-order
@@ -470,13 +477,16 @@ def read_data(header, fh=None, filename=None, index_order='F'):
     return data
 
 
-def read(filename, custom_field_map=None, index_order='F'):
+def read(filename: str, custom_field_map: Optional[NRRDFieldMap] = None, index_order: IndexOrder = 'F') \
+        -> Tuple[npt.NDArray, NRRDHeader]:
     """Read a NRRD file and return the header and data
 
-    See :ref:`user-guide:Reading NRRD files` for more information on reading NRRD files.
+    See :ref:`background/how-to-use:reading nrrd files` for more information on reading NRRD files.
 
     .. note::
-            Users should be aware that the `index_order` argument needs to be consistent between `nrrd.read` and `nrrd.write`. I.e., reading an array with `index_order='F'` will result in a transposed version of the original data and hence the writer needs to be aware of this.
+            Users should be aware that the `index_order` argument needs to be consistent between `nrrd.read` and
+            `nrrd.write`. I.e., reading an array with `index_order='F'` will result in a transposed version of the
+            original data and hence the writer needs to be aware of this.
 
     Parameters
     ----------
@@ -486,9 +496,9 @@ def read(filename, custom_field_map=None, index_order='F'):
         Dictionary used for parsing custom field types where the key is the custom field name and the value is a
         string identifying datatype for the custom field.
     index_order : {'C', 'F'}, optional
-        Specifies the index order of the resulting data array. Either 'C' (C-order) where the dimensions are ordered from
-        slowest-varying to fastest-varying (e.g. (z, y, x)), or 'F' (Fortran-order) where the dimensions are ordered
-        from fastest-varying to slowest-varying (e.g. (x, y, z)).
+        Specifies the index order of the resulting data array. Either 'C' (C-order) where the dimensions are ordered
+        from slowest-varying to fastest-varying (e.g. (z, y, x)), or 'F' (Fortran-order) where the dimensions are
+        ordered from fastest-varying to slowest-varying (e.g. (x, y, z)).
 
     Returns
     -------
@@ -502,7 +512,6 @@ def read(filename, custom_field_map=None, index_order='F'):
     :meth:`write`, :meth:`read_header`, :meth:`read_data`
     """
 
-    """Read a NRRD file and return a tuple (data, header)."""
     with open(filename, 'rb') as fh:
         header = read_header(fh, custom_field_map)
         data = read_data(header, fh, filename, index_order)
